import json
import os
import warnings
from collections import OrderedDict
from pickle import load as pickle_load
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Type,
    Union,
    get_args,
)

from truefoundry.common.utils import (
    get_python_version_major_minor,
    list_pip_packages_installed,
)
from truefoundry.common.warnings import TrueFoundryDeprecationWarning
from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    ModelVersionEnvironment,
    SklearnSerializationFormat,
    XGBoostSerializationFormat,
)
from truefoundry.ml._autogen.entities import artifacts as autogen_artifacts
from truefoundry.ml._autogen.models import infer_signature
from truefoundry.ml.enums import ModelFramework
from truefoundry.ml.log_types.artifacts.utils import (
    get_single_file_path_if_only_one_in_directory,
    to_unix_path,
)
from truefoundry.pydantic_v1 import BaseModel, Field

if TYPE_CHECKING:
    from sklearn.base import BaseEstimator
    from xgboost import Booster, XGBModel

# Map serialization format to corresponding pip packages
SERIALIZATION_FORMAT_TO_PACKAGES_NAME_MAP = {
    SklearnSerializationFormat.JOBLIB: ["joblib"],
    SklearnSerializationFormat.CLOUDPICKLE: ["cloudpickle"],
    XGBoostSerializationFormat.JOBLIB: ["joblib"],
    XGBoostSerializationFormat.CLOUDPICKLE: ["cloudpickle"],
}


class FastAIFramework(autogen_artifacts.FastAIFramework):
    """FastAI model Framework"""

    type: Literal["fastai"] = "fastai"


class GluonFramework(autogen_artifacts.GluonFramework):
    """Gluon model Framework"""

    type: Literal["gluon"] = "gluon"


class H2OFramework(autogen_artifacts.H2OFramework):
    """H2O model Framework"""

    type: Literal["h2o"] = "h2o"


class KerasFramework(autogen_artifacts.KerasFramework):
    """Keras model Framework"""

    type: Literal["keras"] = "keras"


class LightGBMFramework(autogen_artifacts.LightGBMFramework):
    """LightGBM model Framework"""

    type: Literal["lightgbm"] = "lightgbm"


class ONNXFramework(autogen_artifacts.ONNXFramework):
    """ONNX model Framework"""

    type: Literal["onnx"] = "onnx"


class PaddleFramework(autogen_artifacts.PaddleFramework):
    """Paddle model Framework"""

    type: Literal["paddle"] = "paddle"


class PyTorchFramework(autogen_artifacts.PyTorchFramework):
    """PyTorch model Framework"""

    type: Literal["pytorch"] = "pytorch"


class SklearnFramework(autogen_artifacts.SklearnFramework):
    """Sklearn model Framework"""

    type: Literal["sklearn"] = "sklearn"


class SpaCyFramework(autogen_artifacts.SpaCyFramework):
    """SpaCy model Framework"""

    type: Literal["spacy"] = "spacy"


class StatsModelsFramework(autogen_artifacts.StatsModelsFramework):
    """StatsModels model Framework"""

    type: Literal["statsmodels"] = "statsmodels"


class TensorFlowFramework(autogen_artifacts.TensorFlowFramework):
    """TensorFlow model Framework"""

    type: Literal["tensorflow"] = "tensorflow"


class TransformersFramework(autogen_artifacts.TransformersFramework):
    """Transformers model Framework"""

    type: Literal["transformers"] = "transformers"


class XGBoostFramework(autogen_artifacts.XGBoostFramework):
    """XGBoost model Framework"""

    type: Literal["xgboost"] = "xgboost"


# Union of all the model frameworks


ModelFrameworkType = Union[
    FastAIFramework,
    GluonFramework,
    H2OFramework,
    KerasFramework,
    LightGBMFramework,
    ONNXFramework,
    PaddleFramework,
    PyTorchFramework,
    SklearnFramework,
    SpaCyFramework,
    StatsModelsFramework,
    TensorFlowFramework,
    TransformersFramework,
    XGBoostFramework,
]


class _SerializationFormatLoaderRegistry:
    def __init__(self, framework: Type[Union[SklearnFramework, XGBoostFramework]]):
        # An OrderedDict is used to maintain the order of loaders based on priority
        # The loaders are added in the following order:
        #     1. joblib (if available)
        #     2. cloudpickle (if available)
        #     3. pickle (default fallback)
        # This ensures that when looking up a loader, it follows the correct loading priority.
        self._loader_map: Dict[
            Union[SklearnSerializationFormat, XGBoostSerializationFormat],
            Callable[[bytes], object],
        ] = OrderedDict()
        format_class: Union[SklearnSerializationFormat, XGBoostSerializationFormat] = (
            SklearnSerializationFormat
            if framework == SklearnFramework
            else XGBoostSerializationFormat
        )
        is_xgboost = framework == XGBoostFramework

        try:
            from joblib import load as joblib_load

            self._loader_map[format_class.JOBLIB] = joblib_load
        except ImportError:
            pass

        try:
            from cloudpickle import load as cloudpickle_load

            self._loader_map[format_class.CLOUDPICKLE] = cloudpickle_load

        except ImportError:
            pass

        if is_xgboost:
            try:
                from xgboost import Booster

                booster = Booster()
                self._loader_map[format_class.JSON] = booster.load_model
            except ImportError:
                pass

        # Add pickle loader as a fallback
        self._loader_map[format_class.PICKLE] = pickle_load

    def get_loader_map(
        self,
    ) -> Dict[
        Union[SklearnSerializationFormat, XGBoostSerializationFormat],
        Callable[[bytes], object],
    ]:
        return self._loader_map

    def _detect_model_serialization_format(
        self,
        model_file_path: str,
    ) -> Optional[Union[SklearnSerializationFormat, XGBoostSerializationFormat]]:
        """
        The function will attempt to load the model using each different serialization format's loader and return the first successful one.

        Args:
            model_file_path (str): The path to the file to be loaded.

        Returns:
            Optional[Union[SklearnSerializationFormat, XGBoostSerializationFormat]]: The serialization format if successfully loaded, None otherwise.
        """
        # Attempt to load the model using each framework
        for (
            serialization_format,
            loader,
        ) in self._loader_map.items():
            try:
                with open(model_file_path, "rb") as f:
                    loader(f)
                return serialization_format
            except Exception:
                continue
        return None


class _ModelFramework(BaseModel):
    __root__: ModelFrameworkType = Field(discriminator="type")

    @classmethod
    def to_model_framework_type(
        cls,
        framework: Optional[Union[str, ModelFramework, "ModelFrameworkType"]] = None,
    ) -> Optional["ModelFrameworkType"]:
        """
        Converts a ModelFramework or string representation to a ModelFrameworkType object.

        Args:
            framework (Optional[Union[str, ModelFramework, ModelFrameworkType]]): ModelFrameworkType or equivalent input.
                Supported frameworks can be found in `truefoundry.ml.enums.ModelFramework`.
                May be `None` if the framework is unknown or unsupported.
                **Deprecated**: Prefer passing a `ModelFrameworkType` instance.

        Returns:
            ModelFrameworkType corresponding to the input, or None if the input is None.
        """
        if framework is None:
            return None

        # Issue a deprecation warning for str and ModelFramework types
        if isinstance(framework, (str, ModelFramework)):
            warnings.warn(
                "Passing a string or ModelFramework Enum is deprecated. Please use a ModelFrameworkType object.",
                category=TrueFoundryDeprecationWarning,
                stacklevel=2,
            )

        # Convert string to ModelFramework
        if isinstance(framework, str):
            framework = ModelFramework(framework)

        # Convert ModelFramework to ModelFrameworkType
        if isinstance(framework, ModelFramework):
            if framework == ModelFramework.UNKNOWN:
                return None
            return cls.parse_obj({"type": framework.value}).__root__

        # Directly return if already a ModelFrameworkType
        if isinstance(framework, get_args(ModelFrameworkType)):
            return framework

        raise ValueError(
            "framework must be a string, ModelFramework enum, or ModelFrameworkType object"
        )

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[ModelFrameworkType]:
        """Create an instance of ModelFramework from a dict"""
        if obj is None:
            return None

        return cls.parse_obj(obj).__root__


# Mapping of model frameworks to pip packages
_MODEL_FRAMEWORK_TO_PIP_PACKAGES: Dict[Type[ModelFrameworkType], List[str]] = {
    SklearnFramework: ["scikit-learn", "numpy", "pandas"],
    XGBoostFramework: ["xgboost", "numpy", "pandas"],
}


def _get_required_framework_pip_packages(framework: "ModelFrameworkType") -> List[str]:
    """
    Fetches the pip packages required for a given model framework.

    Args:
        framework ("ModelFrameworkType"): The model framework for which to fetch the pip packages.

    Returns:
        List[str]: The list of pip packages required for the given model framework.
                  If no packages are found for the framework type, returns an empty list.
    """
    return _MODEL_FRAMEWORK_TO_PIP_PACKAGES.get(framework.__class__, [])


def _fetch_framework_specific_pip_packages(
    framework: "ModelFrameworkType",
) -> List[str]:
    """
    Fetch the pip packages required for the given framework, including any dependencies
    related to the framework's serialization format.

    Args:
        framework: The framework object (e.g., SklearnFramework, XGBoostFramework).

    Returns:
        List[str]: A list of pip packages for the given framework and environment,
                   including any dependencies based on the serialization format
                   (e.g., ['numpy==1.19.5', ...]).
    """
    framework_package_names = _get_required_framework_pip_packages(framework=framework)

    # Add serialization format dependencies if applicable
    if isinstance(framework, (SklearnFramework, XGBoostFramework)):
        framework_package_names.extend(
            SERIALIZATION_FORMAT_TO_PACKAGES_NAME_MAP.get(
                framework.serialization_format, []
            )
        )
    return [
        f"{package.name}=={package.version}"
        for package in list_pip_packages_installed(
            filter_package_names=framework_package_names
        )
    ]


def auto_update_environment_details(
    environment: ModelVersionEnvironment,
    framework: Optional[ModelFrameworkType],
):
    """
    Auto fetch the environment details if not provided, based on the provided environment and framework.

    Args:
        environment: The environment object that holds environment details like python_version and pip_packages.
        framework: The framework object (e.g., SklearnFramework, XGBoostFramework) that may affect pip_package fetching.
    """
    # Auto fetch python_version if not provided
    if not environment.python_version:
        environment.python_version = get_python_version_major_minor()

    # Framework-specific pip_package handling
    if framework and not environment.pip_packages:
        environment.pip_packages = _fetch_framework_specific_pip_packages(framework)


def _validate_and_get_absolute_model_filepath(
    model_file_or_folder: str,
    model_filepath: Optional[str] = None,
) -> Optional[str]:
    # If no model_filepath is set, resolve it from the directory
    if not model_filepath:
        # If model_filepath is not set, resolve it based on these cases:
        #    - Case 1: model_file_or_folder/model.joblib -> model.joblib
        #    - Case 2: model_file_or_folder/folder/model.joblib -> folder/model.joblib
        #    - Case 3: model_file_or_folder/folder/model.joblib, model_file_or_folder/config.json -> None
        return get_single_file_path_if_only_one_in_directory(model_file_or_folder)

    # If model_filepath is already set, validate and resolve it:
    #    - Case 1: Resolve the absolute file path of the model file relative to the provided directory.
    #        Example: If model_file_or_folder is '/root/models' and model_filepath is 'model.joblib',
    #                 the resolved model file path would be '/root/models/model.joblib'. Validate it.
    #
    #    - Case 2: If model_filepath is a relative path, resolve it to an absolute path based on the provided directory.
    #        Example: If model_file_or_folder is '/root/models' and model_filepath is 'subfolder/model.joblib',
    #                 the resolved path would be '/root/models/subfolder/model.joblib'. Validate it.
    #
    #    - Case 3: Verify that the resolved model file exists and is a valid file.
    #        Example: If the resolved path is '/root/models/model.joblib', check if the file exists.
    #                 If it does not exist, raise a FileNotFoundError.
    #
    #    - Case 4: Ensure the resolved model file is located within the specified directory or is the directory itself.
    #        Example: If the resolved path is '/root/models/model.joblib' and model_file_or_folder is '/root/models',
    #                 the resolved path is valid. If the file lies outside '/root/models', raise a ValueError.
    #

    # If model_filepath is set, Resolve the absolute path of the model file (It can be a relative path or absolute path)
    model_dir = (
        os.path.dirname(model_file_or_folder)
        if os.path.isfile(model_file_or_folder)
        else model_file_or_folder
    )
    absolute_model_filepath = os.path.abspath(os.path.join(model_dir, model_filepath))

    # Validate if resolve valid is within the provided directory or is the same as it
    if not (
        absolute_model_filepath == model_file_or_folder
        or absolute_model_filepath.startswith(model_file_or_folder + os.sep)
    ):
        raise ValueError(
            f"model_filepath '{model_filepath}' must be relative to "
            f"{model_file_or_folder}. Resolved path '{absolute_model_filepath}' is invalid."
        )

    if not os.path.isfile(absolute_model_filepath):
        raise FileNotFoundError(f"Model file not found: {absolute_model_filepath}")

    return absolute_model_filepath


def _validate_and_resolve_model_filepath(
    model_file_or_folder: str,
    model_filepath: Optional[str] = None,
) -> Optional[str]:
    absolute_model_filepath = _validate_and_get_absolute_model_filepath(
        model_file_or_folder=model_file_or_folder, model_filepath=model_filepath
    )
    if absolute_model_filepath:
        return to_unix_path(
            os.path.relpath(absolute_model_filepath, model_file_or_folder)
            if os.path.isdir(model_file_or_folder)
            else os.path.basename(absolute_model_filepath)
        )


def auto_update_model_framework_details(
    framework: "ModelFrameworkType", model_file_or_folder: str
):
    """
    Auto update the model framework details based on the provided model file or folder path.

    Args:
        framework: The framework object (e.g., SklearnFramework, XGBoostFramework) to update.
        model_file_or_folder: The path to the model file or folder.
    """

    # Ensure the model file or folder path is an absolute path
    model_file_or_folder = os.path.abspath(model_file_or_folder)

    if isinstance(framework, (SklearnFramework, XGBoostFramework)):
        framework.model_filepath = _validate_and_resolve_model_filepath(
            model_file_or_folder=model_file_or_folder,
            model_filepath=framework.model_filepath,
        )
        if framework.model_filepath:
            absolute_model_filepath = (
                model_file_or_folder
                if os.path.isfile(model_file_or_folder)
                else os.path.join(model_file_or_folder, framework.model_filepath)
            )
            if not framework.serialization_format:
                loader_registry = _SerializationFormatLoaderRegistry(
                    framework=framework
                )
                framework.serialization_format = (
                    loader_registry._detect_model_serialization_format(
                        model_file_path=absolute_model_filepath
                    )
                )


def _infer_schema(
    model_input: Any,
    model: Union["BaseEstimator", "Booster", "XGBModel"],
    infer_method_name: str = "predict",
) -> Dict[str, Any]:
    if not hasattr(model, infer_method_name):
        raise ValueError(
            f"Model does not have the method '{infer_method_name}' to infer the schema."
        )
    model_infer_method = getattr(model, infer_method_name)
    model_output = model_infer_method(model_input)

    model_signature = infer_signature(
        model_input=model_input, model_output=model_output
    )
    return model_signature.to_dict()


def sklearn_infer_schema(
    model_input: Any,
    model: "BaseEstimator",
    infer_method_name: str = "predict",
) -> autogen_artifacts.SklearnModelSchema:
    """
    Infer the schema of a Sklearn model.

    Args:
        model_input (Any): The input data to be used for schema inference.
        model (Any): The Sklearn model instance.
        infer_method_name (str): The name of the method to be used for schema inference.
            Eg: predict (default), predict_proba
    Returns:
        SklearnModelSchema: The inferred schema of the Sklearn model.
    """
    model_signature_json = _infer_schema(
        model_input=model_input, model=model, infer_method_name=infer_method_name
    )
    return autogen_artifacts.SklearnModelSchema(
        infer_method_name=infer_method_name,
        inputs=json.loads(model_signature_json["inputs"]),
        outputs=json.loads(model_signature_json["outputs"]),
    )


def xgboost_infer_schema(
    model_input: Any,
    model: Union["Booster", "XGBModel"],
    infer_method_name: str = "predict",
) -> autogen_artifacts.XGBoostModelSchema:
    """
    Infer the schema of an XGBoost model.

    Args:
        model_input (Any): The input data to be used for schema inference.
        model (Any): The XGBoost model instance.
        infer_method_name (str): The name of the method to be used for schema inference.
            Eg: predict (default), predict_proba
    Returns:
        XGBoostModelSchema: The inferred schema of the XGBoost model.
    """
    model_signature_json = _infer_schema(
        model_input=model_input, model=model, infer_method_name=infer_method_name
    )
    return autogen_artifacts.XGBoostModelSchema(
        infer_method_name=infer_method_name,
        inputs=json.loads(model_signature_json["inputs"]),
        outputs=json.loads(model_signature_json["outputs"]),
    )



import collections
import tempfile
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    ArtifactType,
)
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.log_types.artifacts.artifact import (
    ArtifactPath,
    ArtifactVersion,
    _log_artifact_version_helper,
)
from truefoundry.ml.log_types.artifacts.utils import (
    _copy_additional_files,
    _validate_artifact_metadata,
    _validate_description,
)
from truefoundry.ml.logger import logger

if TYPE_CHECKING:
    from truefoundry.ml.mlfoundry_run import MlFoundryRun


def _log_artifact_version(
    run: Optional["MlFoundryRun"],
    name: str,
    artifact_paths: List[Union[ArtifactPath, Tuple[str, Optional[str]], Tuple[str]]],
    ml_repo: Optional[str] = None,
    description: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    step: Optional[int] = 0,
    progress: Optional[bool] = None,
) -> ArtifactVersion:
    if not run and not ml_repo:
        raise MlFoundryException("Exactly one of run, ml_repo should be passed")
    for i, artifact_path in enumerate(artifact_paths):
        if isinstance(artifact_path, ArtifactPath):
            continue
        elif isinstance(artifact_path, collections.abc.Sequence) and (
            0 < len(artifact_path) <= 2
        ):
            artifact_paths[i] = ArtifactPath(*artifact_path)
        else:
            raise ValueError(
                "`artifact_path` should be an instance of `truefoundry.ml.ArtifactPath` or a tuple "
                "of (src, dest) path strings"
            )

    metadata = metadata or {}
    step = step or 0  # TODO (chiragjn): remove Optional from step

    _validate_description(description)
    _validate_artifact_metadata(metadata)

    logger.info("Logging the artifact, this might take a while ...")
    temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")

    try:
        temp_dest_to_src_map = _copy_additional_files(
            root_dir=temp_dir.name,
            files_dir="",
            model_dir=None,
            additional_files=artifact_paths,
        )

    # TODO(nikp1172) verify error message when artifact doesn't exist
    except Exception as e:
        temp_dir.cleanup()
        raise MlFoundryException("Failed to log artifact") from e

    return _log_artifact_version_helper(
        run=run,
        ml_repo=ml_repo,
        name=name,
        artifact_type=ArtifactType.ARTIFACT,
        artifact_dir=temp_dir,
        dest_to_src_map=temp_dest_to_src_map,
        description=description,
        internal_metadata=None,
        metadata=metadata,
        step=step,
        progress=progress,
    )



from __future__ import annotations

import threading
from typing import Optional

from truefoundry.common.credential_provider import (
    CredentialProvider,
    EnvCredentialProvider,
    FileCredentialProvider,
)
from truefoundry.common.entities import Token, UserInfo
from truefoundry.common.utils import relogin_error_message
from truefoundry.logger import logger

SESSION_LOCK = threading.RLock()
ACTIVE_SESSION: Optional["Session"] = None


class Session:
    def __init__(self) -> None:
        self._closed = False
        self._cred_provider: Optional[CredentialProvider] = self._get_cred_provider()
        self._user_info: Optional[UserInfo] = self.token.to_user_info()

    @staticmethod
    def _get_cred_provider() -> CredentialProvider:
        final_cred_provider = None
        for cred_provider in [EnvCredentialProvider, FileCredentialProvider]:
            if cred_provider.can_provide():
                final_cred_provider = cred_provider()
                break
        if final_cred_provider is None:
            raise Exception(
                relogin_error_message(
                    "No active session found. Perhaps you are not logged in?",
                )
            )
        return final_cred_provider

    @classmethod
    def new(cls) -> "Session":
        global ACTIVE_SESSION
        with SESSION_LOCK:
            new_session = cls()
            if ACTIVE_SESSION and ACTIVE_SESSION == new_session:
                return ACTIVE_SESSION

            if ACTIVE_SESSION:
                ACTIVE_SESSION.close()

            ACTIVE_SESSION = new_session
            logger.info(
                "Logged in to %r as %r",
                new_session.tfy_host,
                new_session.user_info.user_id,
            )

            return ACTIVE_SESSION

    def close(self):
        self._closed = True
        self._user_info = None
        self._cred_provider = None

    def _assert_not_closed(self):
        if self._closed:
            raise Exception("This session has been deactivated.")

    @property
    def access_token(self) -> str:
        assert self._cred_provider is not None
        return self.token.access_token

    @property
    def token(self) -> Token:
        assert self._cred_provider is not None
        return self._cred_provider.token

    @property
    def tfy_host(self) -> str:
        assert self._cred_provider is not None
        return self._cred_provider.tfy_host

    @property
    def user_info(self) -> UserInfo:
        self._assert_not_closed()
        assert self._user_info is not None
        return self._user_info

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Session):
            return False
        return (
            type(self._cred_provider) == type(other._cred_provider)  # noqa: E721
            and self.user_info == other.user_info
            and self.tfy_host == other.tfy_host
        )



from typing import Optional

from truefoundry.deploy.lib import session


def login(
    host: Optional[str] = None, api_key: Optional[str] = None, relogin: bool = False
):
    session.login(api_key=api_key, relogin=relogin, host=host)



from typing import Any, Dict, Type

import pydantic

PYDANTIC_V2 = pydantic.VERSION.startswith("2.")


def model_dump(
    model: pydantic.BaseModel,
) -> Dict[str, Any]:
    if PYDANTIC_V2:
        return model.model_dump()
    return model.dict()


def model_json_schema(model: Type[pydantic.BaseModel]) -> Dict[str, Any]:
    if PYDANTIC_V2:
        return model.model_json_schema()
    return model.schema()



from truefoundry.pydantic_v1 import BaseModel


class PydanticBase(BaseModel):
    # I can make this a property,
    # but <3.9, it is difficult to access
    # property from classmethod
    @staticmethod
    def get_log_type() -> str:
        raise NotImplementedError()



from __future__ import annotations

import functools
from typing import Callable, TypeVar

import requests
from packaging import version
from typing_extensions import ParamSpec

from truefoundry.common.constants import (
    SERVICEFOUNDRY_CLIENT_MAX_RETRIES,
    VERSION_PREFIX,
)
from truefoundry.common.entities import (
    PythonSDKConfig,
    TenantInfo,
)
from truefoundry.common.request_utils import (
    request_handling,
    requests_retry_session,
)
from truefoundry.common.utils import (
    get_tfy_servers_config,
    timed_lru_cache,
)
from truefoundry.logger import logger
from truefoundry.version import __version__

P = ParamSpec("P")
R = TypeVar("R")


def check_min_cli_version(fn: Callable[P, R]) -> Callable[P, R]:
    @functools.wraps(fn)
    def inner(*args: P.args, **kwargs: P.kwargs) -> R:
        client: "ServiceFoundryServiceClient" = args[0]
        client.check_min_cli_version()
        return fn(*args, **kwargs)

    return inner


def session_with_retries(
    retries: int = SERVICEFOUNDRY_CLIENT_MAX_RETRIES,
) -> requests.Session:
    return requests_retry_session(retries=retries)


@timed_lru_cache(seconds=30 * 60)
def _cached_get_tenant_info(api_server_url: str, tenant_host: str) -> TenantInfo:
    response = session_with_retries().get(
        url=f"{api_server_url}/{VERSION_PREFIX}/tenant-id",
        params={"hostName": tenant_host},
    )
    response_data = request_handling(response)
    return TenantInfo.parse_obj(response_data)


@timed_lru_cache(seconds=30 * 60)
def _cached_get_python_sdk_config(api_server_url: str) -> PythonSDKConfig:
    response = session_with_retries().get(
        url=f"{api_server_url}/{VERSION_PREFIX}/min-cli-version"
    )
    response_data = request_handling(response)
    return PythonSDKConfig.parse_obj(response_data)


class ServiceFoundryServiceClient:
    def __init__(self, tfy_host: str):
        self._tfy_host = tfy_host.strip("/")
        tfy_servers_config = get_tfy_servers_config(self._tfy_host)
        self._tenant_host = tfy_servers_config.tenant_host
        self._api_server_url = tfy_servers_config.servicefoundry_server_url

    @property
    def tfy_host(self) -> str:
        return self._tfy_host

    @property
    def tenant_info(self) -> TenantInfo:
        return _cached_get_tenant_info(
            self._api_server_url,
            tenant_host=self._tenant_host,
        )

    @property
    def python_sdk_config(self) -> PythonSDKConfig:
        return _cached_get_python_sdk_config(
            self._api_server_url,
        )

    @functools.cached_property
    def _min_cli_version_required(self) -> str:
        return self.python_sdk_config.truefoundry_cli_min_version

    def check_min_cli_version(self) -> None:
        if __version__ != "0.0.0":
            # "0.0.0" indicates dev version
            # noinspection PyProtectedMember
            min_cli_version_required = self._min_cli_version_required
            if version.parse(__version__) < version.parse(min_cli_version_required):
                raise Exception(
                    "You are using an outdated version of `truefoundry`.\n"
                    f"Run `pip install truefoundry>={min_cli_version_required}` to install the supported version.",
                )
        else:
            logger.debug("Ignoring minimum cli version check")



import os
from typing import Optional

import rich_click as click

from truefoundry.cli.console import console
from truefoundry.cli.const import COMMAND_CLS
from truefoundry.cli.util import handle_exception_wrapper
from truefoundry.ml._autogen.client import ModelServer  # type: ignore[attr-defined]
from truefoundry.ml.cli.utils import (
    AppName,
    NonEmptyString,
)
from truefoundry.ml.mlfoundry_api import get_client


@click.command(
    name="model",
    cls=COMMAND_CLS,
    help="Generating application code for the specified model version.",
)
@click.option(
    "--name",
    required=True,
    type=AppName(),
    help="Name for the model server deployment",
    show_default=True,
)
@click.option(
    "--model-version-fqn",
    "--model_version_fqn",
    type=NonEmptyString(),
    required=True,
    show_default=True,
    help="Fully Qualified Name (FQN) of the model version to deploy, e.g., 'model:tenant_name/my-model/linear-regression:2'",
)
@click.option(
    "-w",
    "--workspace-fqn",
    "--workspace_fqn",
    type=NonEmptyString(),
    required=True,
    show_default=True,
    help="Fully Qualified Name (FQN) of the workspace to deploy",
)
@click.option(
    "--model-server",
    "--model_server",
    type=click.Choice(ModelServer, case_sensitive=False),
    default=ModelServer.FASTAPI.value,
    show_default=True,
    help="Specify the model server (Case Insensitive).",
)
@click.option(
    "--output-dir",
    "--output_dir",
    type=click.Path(exists=True, file_okay=False, writable=True),
    help="Output directory for the model server code",
    required=False,
    show_default=True,
    default=os.getcwd(),
)
@handle_exception_wrapper
def model_init_command(
    name: str,
    model_version_fqn: str,
    workspace_fqn: str,
    model_server: ModelServer,
    output_dir: Optional[str],
):
    """
    Generates application code for the specified model version.
    """
    ml_client = get_client()
    console.print(f"Generating application code for {model_version_fqn!r}")
    output_dir = ml_client._initialize_model_server(
        name=name,
        model_version_fqn=model_version_fqn,
        workspace_fqn=workspace_fqn,
        model_server=ModelServer[model_server.upper()],
        output_dir=output_dir,
    )
    message = f"""
[bold green]Model Server code initialized successfully![/bold green]

[bold]Code Location:[/bold] {output_dir}

[bold]Next Steps:[/bold]
- Navigate to the model server directory:
[green]cd {output_dir}[/green]
- Refer to the README file in the directory for further instructions.
"""
    console.print(message)


def get_model_init_command():
    return model_init_command



from typing import Callable, Tuple, Union

from flytekit import FlyteContext, PythonFunctionTask
from flytekit.extend import Promise, TaskPlugins

from truefoundry.logger import logger
from truefoundry.workflow import PythonTaskConfig


class TrueFoundryFunctionTask(PythonFunctionTask[PythonTaskConfig]):
    # Note(nikp1172) We are not creating our task type for now, we just re-use their task type for now
    def __init__(
        self,
        task_config: PythonTaskConfig,
        task_function: Callable,
        **kwargs,
    ):
        super().__init__(task_config=task_config, task_function=task_function, **kwargs)

    def local_execute(
        self, ctx: FlyteContext, **kwargs
    ) -> Union[Tuple[Promise], Promise, None]:
        logger.warning(
            "Running pod task locally. Local environment may not match pod environment which may cause issues."
        )
        return super().local_execute(ctx=ctx, **kwargs)

    def get_custom(self, settings) -> dict:
        return {"truefoundry": self._task_config.dict()}


TaskPlugins.register_pythontask_plugin(PythonTaskConfig, TrueFoundryFunctionTask)



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

import json
import re  # noqa: F401

from aenum import Enum


class ArtifactType(str, Enum):
    """
    An enumeration.
    """

    """
    allowed enum values
    """
    ARTIFACT = "artifact"
    MODEL = "model"
    PLOT = "plot"
    IMAGE = "image"
    CHAT_PROMPT = "chat_prompt"
    OPENAPI_MINUS_TOOL = "openapi-tool"
    AGENT = "agent"

    @classmethod
    def from_json(cls, json_str: str) -> ArtifactType:
        """Create an instance of ArtifactType from a JSON string"""
        return ArtifactType(json.loads(json_str))



from truefoundry.deploy.cli.commands.apply_command import get_apply_command
from truefoundry.deploy.cli.commands.build_command import get_build_command
from truefoundry.deploy.cli.commands.delete_command import get_delete_command
from truefoundry.deploy.cli.commands.deploy_command import get_deploy_command
from truefoundry.deploy.cli.commands.deploy_init_command import get_deploy_init_command
from truefoundry.deploy.cli.commands.get_command import get_get_command
from truefoundry.deploy.cli.commands.login_command import get_login_command
from truefoundry.deploy.cli.commands.logout_command import get_logout_command
from truefoundry.deploy.cli.commands.logs_command import get_logs_command
from truefoundry.deploy.cli.commands.patch_application_command import (
    get_patch_application_command,
)
from truefoundry.deploy.cli.commands.patch_command import get_patch_command
from truefoundry.deploy.cli.commands.terminate_comand import get_terminate_command
from truefoundry.deploy.cli.commands.trigger_command import get_trigger_command

__all__ = [
    "get_apply_command",
    "get_build_command",
    "get_delete_command",
    "get_deploy_command",
    "get_deploy_init_command",
    "get_get_command",
    "get_login_command",
    "get_logout_command",
    "get_logs_command",
    "get_patch_application_command",
    "get_patch_command",
    "get_terminate_command",
    "get_trigger_command",
]



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

import re  # noqa: F401
from typing import Optional

from truefoundry.ml._autogen.client.api_client import ApiClient
from truefoundry.ml._autogen.client.api_response import ApiResponse
from truefoundry.ml._autogen.client.exceptions import (  # noqa: F401
    ApiTypeError,
    ApiValueError,
)
from truefoundry.ml._autogen.client.models.create_run_request_dto import (
    CreateRunRequestDto,
)
from truefoundry.ml._autogen.client.models.create_run_response_dto import (
    CreateRunResponseDto,
)
from truefoundry.ml._autogen.client.models.delete_run_request import DeleteRunRequest
from truefoundry.ml._autogen.client.models.delete_tag_request_dto import (
    DeleteTagRequestDto,
)
from truefoundry.ml._autogen.client.models.get_latest_run_log_response_dto import (
    GetLatestRunLogResponseDto,
)
from truefoundry.ml._autogen.client.models.list_latest_run_logs_response_dto import (
    ListLatestRunLogsResponseDto,
)
from truefoundry.ml._autogen.client.models.list_run_logs_response_dto import (
    ListRunLogsResponseDto,
)
from truefoundry.ml._autogen.client.models.log_batch_request_dto import (
    LogBatchRequestDto,
)
from truefoundry.ml._autogen.client.models.log_metric_request_dto import (
    LogMetricRequestDto,
)
from truefoundry.ml._autogen.client.models.log_param_request_dto import (
    LogParamRequestDto,
)
from truefoundry.ml._autogen.client.models.restore_run_request_dto import (
    RestoreRunRequestDto,
)
from truefoundry.ml._autogen.client.models.run_response_dto import RunResponseDto
from truefoundry.ml._autogen.client.models.search_runs_request_dto import (
    SearchRunsRequestDto,
)
from truefoundry.ml._autogen.client.models.search_runs_response_dto import (
    SearchRunsResponseDto,
)
from truefoundry.ml._autogen.client.models.set_tag_request_dto import SetTagRequestDto
from truefoundry.ml._autogen.client.models.store_run_logs_request_dto import (
    StoreRunLogsRequestDto,
)
from truefoundry.ml._autogen.client.models.update_run_request_dto import (
    UpdateRunRequestDto,
)
from truefoundry.ml._autogen.client.models.update_run_response_dto import (
    UpdateRunResponseDto,
)
from truefoundry.pydantic_v1 import (
    StrictInt,
    StrictStr,
    conlist,
    validate_arguments,
)


class RunsApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_arguments
    def create_run_post(
        self, create_run_request_dto: CreateRunRequestDto, **kwargs
    ) -> CreateRunResponseDto:  # noqa: E501
        """Create Run  # noqa: E501

        Create Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.create_run_post(create_run_request_dto, async_req=True)
        >>> result = thread.get()

        :param create_run_request_dto: (required)
        :type create_run_request_dto: CreateRunRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: CreateRunResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the create_run_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.create_run_post_with_http_info(create_run_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def create_run_post_with_http_info(
        self, create_run_request_dto: CreateRunRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Create Run  # noqa: E501

        Create Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.create_run_post_with_http_info(create_run_request_dto, async_req=True)
        >>> result = thread.get()

        :param create_run_request_dto: (required)
        :type create_run_request_dto: CreateRunRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(CreateRunResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["create_run_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method create_run_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["create_run_request_dto"] is not None:
            _body_params = _params["create_run_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "CreateRunResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/create",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def delete_run_post(self, delete_run_request: DeleteRunRequest, **kwargs) -> object:  # noqa: E501
        """Delete Run  # noqa: E501

        Delete Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.delete_run_post(delete_run_request, async_req=True)
        >>> result = thread.get()

        :param delete_run_request: (required)
        :type delete_run_request: DeleteRunRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: object
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the delete_run_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.delete_run_post_with_http_info(delete_run_request, **kwargs)  # noqa: E501

    @validate_arguments
    def delete_run_post_with_http_info(
        self, delete_run_request: DeleteRunRequest, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Delete Run  # noqa: E501

        Delete Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.delete_run_post_with_http_info(delete_run_request, async_req=True)
        >>> result = thread.get()

        :param delete_run_request: (required)
        :type delete_run_request: DeleteRunRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(object, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["delete_run_request"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method delete_run_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["delete_run_request"] is not None:
            _body_params = _params["delete_run_request"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/delete",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def delete_tag_post(
        self, delete_tag_request_dto: DeleteTagRequestDto, **kwargs
    ) -> object:  # noqa: E501
        """Delete Tag  # noqa: E501

        Delete Tag  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.delete_tag_post(delete_tag_request_dto, async_req=True)
        >>> result = thread.get()

        :param delete_tag_request_dto: (required)
        :type delete_tag_request_dto: DeleteTagRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: object
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the delete_tag_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.delete_tag_post_with_http_info(delete_tag_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def delete_tag_post_with_http_info(
        self, delete_tag_request_dto: DeleteTagRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Delete Tag  # noqa: E501

        Delete Tag  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.delete_tag_post_with_http_info(delete_tag_request_dto, async_req=True)
        >>> result = thread.get()

        :param delete_tag_request_dto: (required)
        :type delete_tag_request_dto: DeleteTagRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(object, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["delete_tag_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method delete_tag_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["delete_tag_request_dto"] is not None:
            _body_params = _params["delete_tag_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/delete-tag",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def get_run_by_fqn_get(self, run_fqn: StrictStr, **kwargs) -> RunResponseDto:  # noqa: E501
        """Get Run By Fqn  # noqa: E501

        Get Run by run FQN  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_run_by_fqn_get(run_fqn, async_req=True)
        >>> result = thread.get()

        :param run_fqn: (required)
        :type run_fqn: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: RunResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the get_run_by_fqn_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.get_run_by_fqn_get_with_http_info(run_fqn, **kwargs)  # noqa: E501

    @validate_arguments
    def get_run_by_fqn_get_with_http_info(
        self, run_fqn: StrictStr, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Get Run By Fqn  # noqa: E501

        Get Run by run FQN  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_run_by_fqn_get_with_http_info(run_fqn, async_req=True)
        >>> result = thread.get()

        :param run_fqn: (required)
        :type run_fqn: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(RunResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["run_fqn"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_run_by_fqn_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("run_fqn") is not None:  # noqa: E501
            _query_params.append(("run_fqn", _params["run_fqn"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "RunResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/get-by-fqn",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def get_run_by_name_get(
        self,
        run_name: StrictStr,
        experiment_id: Optional[StrictStr] = None,
        experiment_name: Optional[StrictStr] = None,
        **kwargs,
    ) -> RunResponseDto:  # noqa: E501
        """Get Run By Name  # noqa: E501

        Get Run by run name  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_run_by_name_get(run_name, experiment_id, experiment_name, async_req=True)
        >>> result = thread.get()

        :param run_name: (required)
        :type run_name: str
        :param experiment_id:
        :type experiment_id: str
        :param experiment_name:
        :type experiment_name: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: RunResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the get_run_by_name_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.get_run_by_name_get_with_http_info(
            run_name, experiment_id, experiment_name, **kwargs
        )  # noqa: E501

    @validate_arguments
    def get_run_by_name_get_with_http_info(
        self,
        run_name: StrictStr,
        experiment_id: Optional[StrictStr] = None,
        experiment_name: Optional[StrictStr] = None,
        **kwargs,
    ) -> ApiResponse:  # noqa: E501
        """Get Run By Name  # noqa: E501

        Get Run by run name  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_run_by_name_get_with_http_info(run_name, experiment_id, experiment_name, async_req=True)
        >>> result = thread.get()

        :param run_name: (required)
        :type run_name: str
        :param experiment_id:
        :type experiment_id: str
        :param experiment_name:
        :type experiment_name: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(RunResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["run_name", "experiment_id", "experiment_name"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_run_by_name_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("run_name") is not None:  # noqa: E501
            _query_params.append(("run_name", _params["run_name"]))

        if _params.get("experiment_id") is not None:  # noqa: E501
            _query_params.append(("experiment_id", _params["experiment_id"]))

        if _params.get("experiment_name") is not None:  # noqa: E501
            _query_params.append(("experiment_name", _params["experiment_name"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "RunResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/get-by-name",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def get_run_get(
        self,
        run_id: Optional[StrictStr] = None,
        run_uuid: Optional[StrictStr] = None,
        **kwargs,
    ) -> RunResponseDto:  # noqa: E501
        """Get Run  # noqa: E501

        Get Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_run_get(run_id, run_uuid, async_req=True)
        >>> result = thread.get()

        :param run_id:
        :type run_id: str
        :param run_uuid:
        :type run_uuid: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: RunResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the get_run_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.get_run_get_with_http_info(run_id, run_uuid, **kwargs)  # noqa: E501

    @validate_arguments
    def get_run_get_with_http_info(
        self,
        run_id: Optional[StrictStr] = None,
        run_uuid: Optional[StrictStr] = None,
        **kwargs,
    ) -> ApiResponse:  # noqa: E501
        """Get Run  # noqa: E501

        Get Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_run_get_with_http_info(run_id, run_uuid, async_req=True)
        >>> result = thread.get()

        :param run_id:
        :type run_id: str
        :param run_uuid:
        :type run_uuid: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(RunResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["run_id", "run_uuid"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_run_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("run_id") is not None:  # noqa: E501
            _query_params.append(("run_id", _params["run_id"]))

        if _params.get("run_uuid") is not None:  # noqa: E501
            _query_params.append(("run_uuid", _params["run_uuid"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "RunResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/get",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def hard_delete_run_post(
        self, delete_run_request: DeleteRunRequest, **kwargs
    ) -> object:  # noqa: E501
        """Hard Delete Run  # noqa: E501

        Hard Delete Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.hard_delete_run_post(delete_run_request, async_req=True)
        >>> result = thread.get()

        :param delete_run_request: (required)
        :type delete_run_request: DeleteRunRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: object
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the hard_delete_run_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.hard_delete_run_post_with_http_info(delete_run_request, **kwargs)  # noqa: E501

    @validate_arguments
    def hard_delete_run_post_with_http_info(
        self, delete_run_request: DeleteRunRequest, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Hard Delete Run  # noqa: E501

        Hard Delete Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.hard_delete_run_post_with_http_info(delete_run_request, async_req=True)
        >>> result = thread.get()

        :param delete_run_request: (required)
        :type delete_run_request: DeleteRunRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(object, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["delete_run_request"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method hard_delete_run_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["delete_run_request"] is not None:
            _body_params = _params["delete_run_request"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/hard-delete",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def latest_run_log_get(
        self, run_uuid: StrictStr, key: StrictStr, log_type: StrictStr, **kwargs
    ) -> GetLatestRunLogResponseDto:  # noqa: E501
        """Latest Run Log  # noqa: E501

        Get latest run log for a key  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.latest_run_log_get(run_uuid, key, log_type, async_req=True)
        >>> result = thread.get()

        :param run_uuid: (required)
        :type run_uuid: str
        :param key: (required)
        :type key: str
        :param log_type: (required)
        :type log_type: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: GetLatestRunLogResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the latest_run_log_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.latest_run_log_get_with_http_info(run_uuid, key, log_type, **kwargs)  # noqa: E501

    @validate_arguments
    def latest_run_log_get_with_http_info(
        self, run_uuid: StrictStr, key: StrictStr, log_type: StrictStr, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Latest Run Log  # noqa: E501

        Get latest run log for a key  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.latest_run_log_get_with_http_info(run_uuid, key, log_type, async_req=True)
        >>> result = thread.get()

        :param run_uuid: (required)
        :type run_uuid: str
        :param key: (required)
        :type key: str
        :param log_type: (required)
        :type log_type: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(GetLatestRunLogResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["run_uuid", "key", "log_type"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method latest_run_log_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("run_uuid") is not None:  # noqa: E501
            _query_params.append(("run_uuid", _params["run_uuid"]))

        if _params.get("key") is not None:  # noqa: E501
            _query_params.append(("key", _params["key"]))

        if _params.get("log_type") is not None:  # noqa: E501
            _query_params.append(("log_type", _params["log_type"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "GetLatestRunLogResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/run-logs/latest-run-log",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def list_latest_run_logs_get(
        self,
        run_uuid: StrictStr,
        key: Optional[StrictStr] = None,
        log_type: Optional[StrictStr] = None,
        **kwargs,
    ) -> ListLatestRunLogsResponseDto:  # noqa: E501
        """List Latest Run Logs  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_latest_run_logs_get(run_uuid, key, log_type, async_req=True)
        >>> result = thread.get()

        :param run_uuid: (required)
        :type run_uuid: str
        :param key:
        :type key: str
        :param log_type:
        :type log_type: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ListLatestRunLogsResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the list_latest_run_logs_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.list_latest_run_logs_get_with_http_info(
            run_uuid, key, log_type, **kwargs
        )  # noqa: E501

    @validate_arguments
    def list_latest_run_logs_get_with_http_info(
        self,
        run_uuid: StrictStr,
        key: Optional[StrictStr] = None,
        log_type: Optional[StrictStr] = None,
        **kwargs,
    ) -> ApiResponse:  # noqa: E501
        """List Latest Run Logs  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_latest_run_logs_get_with_http_info(run_uuid, key, log_type, async_req=True)
        >>> result = thread.get()

        :param run_uuid: (required)
        :type run_uuid: str
        :param key:
        :type key: str
        :param log_type:
        :type log_type: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(ListLatestRunLogsResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["run_uuid", "key", "log_type"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method list_latest_run_logs_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("run_uuid") is not None:  # noqa: E501
            _query_params.append(("run_uuid", _params["run_uuid"]))

        if _params.get("key") is not None:  # noqa: E501
            _query_params.append(("key", _params["key"]))

        if _params.get("log_type") is not None:  # noqa: E501
            _query_params.append(("log_type", _params["log_type"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "ListLatestRunLogsResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/run-logs/list-latest",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def list_run_logs_get(
        self,
        run_uuid: StrictStr,
        key: Optional[StrictStr] = None,
        log_type: Optional[StrictStr] = None,
        request_body: Optional[conlist(StrictInt)] = None,
        **kwargs,
    ) -> ListRunLogsResponseDto:  # noqa: E501
        """List Run Logs  # noqa: E501

        List Run Logs  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_run_logs_get(run_uuid, key, log_type, request_body, async_req=True)
        >>> result = thread.get()

        :param run_uuid: (required)
        :type run_uuid: str
        :param key:
        :type key: str
        :param log_type:
        :type log_type: str
        :param request_body:
        :type request_body: List[int]
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ListRunLogsResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the list_run_logs_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.list_run_logs_get_with_http_info(
            run_uuid, key, log_type, request_body, **kwargs
        )  # noqa: E501

    @validate_arguments
    def list_run_logs_get_with_http_info(
        self,
        run_uuid: StrictStr,
        key: Optional[StrictStr] = None,
        log_type: Optional[StrictStr] = None,
        request_body: Optional[conlist(StrictInt)] = None,
        **kwargs,
    ) -> ApiResponse:  # noqa: E501
        """List Run Logs  # noqa: E501

        List Run Logs  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_run_logs_get_with_http_info(run_uuid, key, log_type, request_body, async_req=True)
        >>> result = thread.get()

        :param run_uuid: (required)
        :type run_uuid: str
        :param key:
        :type key: str
        :param log_type:
        :type log_type: str
        :param request_body:
        :type request_body: List[int]
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(ListRunLogsResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["run_uuid", "key", "log_type", "request_body"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method list_run_logs_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("run_uuid") is not None:  # noqa: E501
            _query_params.append(("run_uuid", _params["run_uuid"]))

        if _params.get("key") is not None:  # noqa: E501
            _query_params.append(("key", _params["key"]))

        if _params.get("log_type") is not None:  # noqa: E501
            _query_params.append(("log_type", _params["log_type"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["request_body"] is not None:
            _body_params = _params["request_body"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "ListRunLogsResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/run-logs/list",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def log_metric_post(
        self, log_metric_request_dto: LogMetricRequestDto, **kwargs
    ) -> object:  # noqa: E501
        """Log Metric  # noqa: E501

        Log Metric  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.log_metric_post(log_metric_request_dto, async_req=True)
        >>> result = thread.get()

        :param log_metric_request_dto: (required)
        :type log_metric_request_dto: LogMetricRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: object
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the log_metric_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.log_metric_post_with_http_info(log_metric_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def log_metric_post_with_http_info(
        self, log_metric_request_dto: LogMetricRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Log Metric  # noqa: E501

        Log Metric  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.log_metric_post_with_http_info(log_metric_request_dto, async_req=True)
        >>> result = thread.get()

        :param log_metric_request_dto: (required)
        :type log_metric_request_dto: LogMetricRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(object, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["log_metric_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method log_metric_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["log_metric_request_dto"] is not None:
            _body_params = _params["log_metric_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/log-metric",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def log_parameter_post(
        self, log_param_request_dto: LogParamRequestDto, **kwargs
    ) -> object:  # noqa: E501
        """Log Parameter  # noqa: E501

        Log Param  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.log_parameter_post(log_param_request_dto, async_req=True)
        >>> result = thread.get()

        :param log_param_request_dto: (required)
        :type log_param_request_dto: LogParamRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: object
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the log_parameter_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.log_parameter_post_with_http_info(log_param_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def log_parameter_post_with_http_info(
        self, log_param_request_dto: LogParamRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Log Parameter  # noqa: E501

        Log Param  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.log_parameter_post_with_http_info(log_param_request_dto, async_req=True)
        >>> result = thread.get()

        :param log_param_request_dto: (required)
        :type log_param_request_dto: LogParamRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(object, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["log_param_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method log_parameter_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["log_param_request_dto"] is not None:
            _body_params = _params["log_param_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/log-parameter",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def log_run_batch_post(
        self, log_batch_request_dto: LogBatchRequestDto, **kwargs
    ) -> object:  # noqa: E501
        """Log Run Batch  # noqa: E501

        Log Batch  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.log_run_batch_post(log_batch_request_dto, async_req=True)
        >>> result = thread.get()

        :param log_batch_request_dto: (required)
        :type log_batch_request_dto: LogBatchRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: object
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the log_run_batch_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.log_run_batch_post_with_http_info(log_batch_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def log_run_batch_post_with_http_info(
        self, log_batch_request_dto: LogBatchRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Log Run Batch  # noqa: E501

        Log Batch  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.log_run_batch_post_with_http_info(log_batch_request_dto, async_req=True)
        >>> result = thread.get()

        :param log_batch_request_dto: (required)
        :type log_batch_request_dto: LogBatchRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(object, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["log_batch_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method log_run_batch_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["log_batch_request_dto"] is not None:
            _body_params = _params["log_batch_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/log-batch",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def restore_run_post(
        self, restore_run_request_dto: RestoreRunRequestDto, **kwargs
    ) -> object:  # noqa: E501
        """Restore Run  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.restore_run_post(restore_run_request_dto, async_req=True)
        >>> result = thread.get()

        :param restore_run_request_dto: (required)
        :type restore_run_request_dto: RestoreRunRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: object
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the restore_run_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.restore_run_post_with_http_info(restore_run_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def restore_run_post_with_http_info(
        self, restore_run_request_dto: RestoreRunRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Restore Run  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.restore_run_post_with_http_info(restore_run_request_dto, async_req=True)
        >>> result = thread.get()

        :param restore_run_request_dto: (required)
        :type restore_run_request_dto: RestoreRunRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(object, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["restore_run_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method restore_run_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["restore_run_request_dto"] is not None:
            _body_params = _params["restore_run_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/restore",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def run_logs_post(
        self, store_run_logs_request_dto: StoreRunLogsRequestDto, **kwargs
    ) -> object:  # noqa: E501
        """Run Logs  # noqa: E501

        Store Run Logs  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.run_logs_post(store_run_logs_request_dto, async_req=True)
        >>> result = thread.get()

        :param store_run_logs_request_dto: (required)
        :type store_run_logs_request_dto: StoreRunLogsRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: object
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the run_logs_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.run_logs_post_with_http_info(store_run_logs_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def run_logs_post_with_http_info(
        self, store_run_logs_request_dto: StoreRunLogsRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Run Logs  # noqa: E501

        Store Run Logs  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.run_logs_post_with_http_info(store_run_logs_request_dto, async_req=True)
        >>> result = thread.get()

        :param store_run_logs_request_dto: (required)
        :type store_run_logs_request_dto: StoreRunLogsRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(object, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["store_run_logs_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method run_logs_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["store_run_logs_request_dto"] is not None:
            _body_params = _params["store_run_logs_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/run-logs",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def search_runs_post(
        self, search_runs_request_dto: SearchRunsRequestDto, **kwargs
    ) -> SearchRunsResponseDto:  # noqa: E501
        """Search Runs  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.search_runs_post(search_runs_request_dto, async_req=True)
        >>> result = thread.get()

        :param search_runs_request_dto: (required)
        :type search_runs_request_dto: SearchRunsRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: SearchRunsResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the search_runs_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.search_runs_post_with_http_info(search_runs_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def search_runs_post_with_http_info(
        self, search_runs_request_dto: SearchRunsRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Search Runs  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.search_runs_post_with_http_info(search_runs_request_dto, async_req=True)
        >>> result = thread.get()

        :param search_runs_request_dto: (required)
        :type search_runs_request_dto: SearchRunsRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(SearchRunsResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["search_runs_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method search_runs_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["search_runs_request_dto"] is not None:
            _body_params = _params["search_runs_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "SearchRunsResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/search",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def set_tag_request_post(
        self, set_tag_request_dto: SetTagRequestDto, **kwargs
    ) -> object:  # noqa: E501
        """Set Tag Request  # noqa: E501

        Set Tag  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.set_tag_request_post(set_tag_request_dto, async_req=True)
        >>> result = thread.get()

        :param set_tag_request_dto: (required)
        :type set_tag_request_dto: SetTagRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: object
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the set_tag_request_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.set_tag_request_post_with_http_info(set_tag_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def set_tag_request_post_with_http_info(
        self, set_tag_request_dto: SetTagRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Set Tag Request  # noqa: E501

        Set Tag  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.set_tag_request_post_with_http_info(set_tag_request_dto, async_req=True)
        >>> result = thread.get()

        :param set_tag_request_dto: (required)
        :type set_tag_request_dto: SetTagRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(object, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["set_tag_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method set_tag_request_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["set_tag_request_dto"] is not None:
            _body_params = _params["set_tag_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "object",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/set-tag",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def update_run_post(
        self, update_run_request_dto: UpdateRunRequestDto, **kwargs
    ) -> UpdateRunResponseDto:  # noqa: E501
        """Update Run  # noqa: E501

        Update Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.update_run_post(update_run_request_dto, async_req=True)
        >>> result = thread.get()

        :param update_run_request_dto: (required)
        :type update_run_request_dto: UpdateRunRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UpdateRunResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the update_run_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.update_run_post_with_http_info(update_run_request_dto, **kwargs)  # noqa: E501

    @validate_arguments
    def update_run_post_with_http_info(
        self, update_run_request_dto: UpdateRunRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Update Run  # noqa: E501

        Update Run  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.update_run_post_with_http_info(update_run_request_dto, async_req=True)
        >>> result = thread.get()

        :param update_run_request_dto: (required)
        :type update_run_request_dto: UpdateRunRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(UpdateRunResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["update_run_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method update_run_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["update_run_request_dto"] is not None:
            _body_params = _params["update_run_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "UpdateRunResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/runs/update",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

import json
import re  # noqa: F401

from aenum import Enum


class Method(str, Enum):
    """
    HTTP Method for the tool call HTTP request
    """

    """
    allowed enum values
    """
    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"
    PATCH = "patch"

    @classmethod
    def from_json(cls, json_str: str) -> Method:
        """Create an instance of Method from a JSON string"""
        return Method(json.loads(json_str))



from typing import Optional


# TODO (chiragjn): We need to establish uniform exception handling across codebase
class HttpRequestException(Exception):
    def __init__(self, message: str, status_code: Optional[int] = None):
        self.message = str(message)
        self.status_code = status_code
        super().__init__(message)

    def __str__(self) -> str:
        return self.message or ""


class BadRequestException(HttpRequestException):
    pass



from truefoundry_sdk import (
    AssistantMessage,
    ChatPromptManifest,
    DeveloperMessage,
    FunctionSchema,
    ModelConfiguration,
    Parameters,
    PromptVersion,
    SystemMessage,
    ToolCall,
    ToolMessage,
    ToolSchema,
    UserMessage,
)
from truefoundry_sdk.client import TrueFoundry

from truefoundry._client import client
from truefoundry.common.warnings import (
    suppress_truefoundry_deprecation_warnings,
    suppress_truefoundry_ml_autogen_warnings,
    surface_truefoundry_deprecation_warnings,
)
from truefoundry.deploy.core import login, logout
from truefoundry.ml.prompt_utils import render_prompt

suppress_truefoundry_ml_autogen_warnings()
surface_truefoundry_deprecation_warnings()

__all__ = [
    "AssistantMessage",
    "ChatPromptManifest",
    "client",
    "DeveloperMessage",
    "FunctionSchema",
    "login",
    "logout",
    "ModelConfiguration",
    "Parameters",
    "PromptVersion",
    "render_prompt",
    "suppress_truefoundry_deprecation_warnings",
    "SystemMessage",
    "TrueFoundry",
    "ToolCall",
    "ToolMessage",
    "ToolSchema",
    "UserMessage",
]



import json
from contextlib import contextmanager
from typing import Any, Dict, Optional, Type, Union

import requests
from requests import Response
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from truefoundry.common.exceptions import BadRequestException, HttpRequestException


# TODO (chiragjn): Rename this to json_response_handling
def request_handling(response: Response) -> Optional[Any]:
    status_code = response.status_code
    if 200 <= status_code <= 299:
        if response.content == b"":
            # TODO (chiragjn): Do we really need this empty check?
            return None
        try:
            return response.json()
        except json.JSONDecodeError as je:
            raise ValueError(
                f"Failed to parse response as json. Response: {response.text}",
            ) from je

    try:
        message = str(response.json()["message"])
    except Exception:
        message = response.text

    if 400 <= status_code <= 499:
        raise BadRequestException(message=message, status_code=response.status_code)
    if 500 <= status_code <= 599:
        raise HttpRequestException(message=message, status_code=response.status_code)

    raise HttpRequestException("Unknown error occurred", status_code=status_code)


def urllib3_retry(
    retries: int = 2,
    backoff_factor: float = 0.3,
    status_forcelist=(408, 429, 500, 502, 503, 504, 524),
    method_whitelist=frozenset({"GET", "POST"}),
    raise_on_status: bool = False,
):
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        status=retries,
        backoff_factor=backoff_factor,
        allowed_methods=method_whitelist,
        status_forcelist=status_forcelist,
        respect_retry_after_header=True,
        raise_on_status=raise_on_status,
    )
    return retry


def requests_retry_session(
    retries: int = 2,
    backoff_factor: float = 0.3,
    status_forcelist=(408, 429, 417, 500, 502, 503, 504, 524),
    method_whitelist=frozenset({"GET", "POST"}),
    raise_on_status: bool = False,
    session: Optional[requests.Session] = None,
) -> requests.Session:
    """
    Returns a `requests` session with retry capabilities for certain HTTP status codes.

    Args:
        retries (int): The number of retries for HTTP requests.
        backoff_factor (float): The backoff factor for exponential backoff during retries.
        status_forcelist (tuple): A tuple of HTTP status codes that should trigger a retry.
        method_whitelist (frozenset): The set of HTTP methods that should be retried.
        session (requests.Session, optional): An optional existing requests session to use.

    Returns:
        requests.Session: A session with retry capabilities.
    """
    # Implementation taken from https://www.peterbe.com/plog/best-practice-with-retries-with-requests
    session = session or requests.Session()
    retry = urllib3_retry(
        retries=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
        method_whitelist=method_whitelist,
        raise_on_status=raise_on_status,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def http_request(
    *,
    method: str,
    url: str,
    token: Optional[str] = None,
    timeout=None,
    headers: Optional[Dict[str, str]] = None,
    session: Optional[requests.Session] = None,
    exception_class: Type[HttpRequestException] = HttpRequestException,
    **kwargs,
) -> requests.Response:
    session = session or requests.Session()
    headers = headers or {}
    if token is not None:
        headers["Authorization"] = f"Bearer {token}"
    try:
        response = session.request(
            method=method, url=url, headers=headers, timeout=timeout, **kwargs
        )
        return response
    except requests.exceptions.ConnectionError as ce:
        raise exception_class("Failed to connect to TrueFoundry") from ce
    except requests.exceptions.Timeout as te:
        raise exception_class(f"Request to {url} timed out") from te
    except Exception as e:
        raise exception_class(f"Request to {url} failed with error {str(e)}") from e


@contextmanager
def cloud_storage_http_request(
    *,
    method: str,
    url: str,
    session: Optional[requests.Session] = None,
    timeout: Optional[Union[int, float]] = None,
    exception_class: Type[HttpRequestException] = HttpRequestException,
    **kwargs,
):
    """
    Performs an HTTP PUT/GET request using Python's `requests` module with automatic retry.
    """
    # Note: This does not support auth and is only meant to be used for pre-signed URLs.
    session = session or requests_retry_session(retries=5, backoff_factor=0.5)
    headers = kwargs.get("headers", {}) or {}
    if "blob.core.windows.net" in url:
        headers.update({"x-ms-blob-type": "BlockBlob"})
    if method.lower() not in ("put", "get"):
        raise ValueError(f"Illegal http method: {method}")
    yield http_request(
        method=method,
        url=url,
        session=session,
        timeout=timeout,
        headers=headers,
        exception_class=exception_class,
        **kwargs,
    )


# TODO: Try and eliminate this
def augmented_raise_for_status(
    response, exception_class: Type[HttpRequestException] = HttpRequestException
):
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as he:
        raise exception_class(
            message=f"Request failed with status code {he.response.status_code}. Response: {he.response.text}",
            status_code=he.response.status_code,
        ) from he



# file: tfy_signed_url_fs.py
# pylint: disable=W0223
import io
import os
from concurrent.futures import FIRST_EXCEPTION, Future, ThreadPoolExecutor, wait
from pathlib import Path
from threading import Event
from typing import List, Optional, Tuple

from fsspec.spec import DEFAULT_CALLBACK, AbstractBufferedFile, AbstractFileSystem

from truefoundry.common.constants import ENV_VARS
from truefoundry.common.storage_provider_utils import (
    MultiPartUpload,
    SignedURL,
    _FileMultiPartInfo,
    decide_file_parts,
    s3_compatible_multipart_upload,
)
from truefoundry.workflow.remote_filesystem.logger import log_time, logger
from truefoundry.workflow.remote_filesystem.tfy_signed_url_client import (
    LOG_PREFIX,
    SignedURLClient,
    SignedURLMultipartUploadAPIResponseDto,
)

MULTIPART_SUPPORTED_PROVIDERS = ["s3"]


def _add_file_for_upload(
    local_file: str,
    remote_path: str,
    files_for_normal_upload: List[Tuple[str, str, _FileMultiPartInfo]],
    files_for_multipart_upload: List[Tuple[str, str, _FileMultiPartInfo]],
    multipart_upload_allowed: bool,
):
    multipart_info = decide_file_parts(local_file, multipart_upload_allowed)
    if multipart_info.num_parts == 1:
        files_for_normal_upload.append((remote_path, local_file, multipart_info))
    else:
        files_for_multipart_upload.append((remote_path, local_file, multipart_info))


class SignedURLFileSystem(AbstractFileSystem):
    def __init__(
        self, base_url: Optional[str] = None, token: Optional[str] = None, **kwargs
    ):
        super().__init__()
        base_url = base_url or ENV_VARS.TFY_INTERNAL_SIGNED_URL_SERVER_HOST
        token = token or ENV_VARS.TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN
        self.client = SignedURLClient(base_url, token)

    @log_time(prefix=LOG_PREFIX)
    def exists(self, path, **kwargs):
        """Check if a file exists at the given path."""
        return self.client.exists(path)

    @log_time(prefix=LOG_PREFIX)
    def get(
        self,
        rpath,
        lpath,
        recursive=False,
        callback=DEFAULT_CALLBACK,
        maxdepth=None,
        **kwargs,
    ):
        """Get file(s) to local"""
        # TODO: Add support for ThreadPoolExecutor here
        # TODO: Add support for async download
        # TODO: Do a proper error handling here
        if self.isdir(rpath):
            if not recursive:
                raise ValueError(
                    f"{rpath} is a directory, but recursive is not enabled."
                )

            # Handle recursive download
            files = self.ls(rpath, detail=True)

            for file_info in files:
                file_path = file_info.path.rstrip("/").rsplit("/")[-1]

                is_directory = file_info.is_directory
                # Construct the relative path for local download
                relative_path = rpath.rstrip("/") + "/" + file_path
                target_local_path = lpath.rstrip("/") + "/" + file_path

                if is_directory:
                    # If it's a directory, create the directory locally
                    Path(target_local_path).mkdir(parents=True, exist_ok=True)
                    relative_path = relative_path + "/"
                    if recursive:
                        # Recursively download the contents of the directory
                        self.get(
                            relative_path,
                            target_local_path,
                            recursive=True,
                            maxdepth=maxdepth,
                            **kwargs,
                        )
                else:
                    self.client.download(
                        storage_uri=relative_path, local_path=target_local_path
                    )

        else:
            # Ensure the directory exists first
            target_local_path = lpath
            if target_local_path.endswith("/"):
                # If it ends with "/", it's a directory, so create the directory first
                target_local_path = os.path.join(
                    target_local_path, rpath.rsplit("/", 1)[-1]
                )
            # Create the directory for the target file path (common for both cases)
            Path(os.path.dirname(target_local_path)).mkdir(parents=True, exist_ok=True)
            self.client.download(storage_uri=rpath, local_path=target_local_path)

    @log_time(prefix=LOG_PREFIX)
    def put(
        self,
        lpath,
        rpath,
        recursive=False,
        callback=DEFAULT_CALLBACK,
        maxdepth=None,
        **kwargs,
    ):
        files_for_normal_upload: List[Tuple[str, str, _FileMultiPartInfo]] = []
        files_for_multipart_upload: List[Tuple[str, str, _FileMultiPartInfo]] = []
        local_path = Path(lpath)
        multipart_upload_allowed = (
            self.protocol in MULTIPART_SUPPORTED_PROVIDERS
            and not ENV_VARS.TFY_ARTIFACTS_DISABLE_MULTIPART_UPLOAD
        )
        if local_path.is_dir():
            if not recursive:
                raise ValueError(
                    f"{lpath} is a directory, but recursive is set to False."
                )

            # Optionally limit recursion depth
            max_depth = maxdepth if maxdepth is not None else float("inf")

            # Walk through the directory structure
            for root, _, files in os.walk(lpath):
                current_depth = Path(root).relative_to(local_path).parts
                if len(current_depth) > max_depth:
                    continue  # Skip files deeper than the max depth

                rel_dir = Path(root).relative_to(local_path)
                remote_dir = (
                    rpath.rstrip("/")
                    if rel_dir == Path(".")
                    else rpath.rstrip("/") + "/" + str(rel_dir)
                )

                # Upload each file
                for file in files:
                    local_file_path = Path(root) / file
                    remote_file_path = f"{remote_dir}/{file}"
                    _add_file_for_upload(
                        local_file=str(local_file_path),
                        remote_path=str(remote_file_path),
                        files_for_normal_upload=files_for_normal_upload,
                        files_for_multipart_upload=files_for_multipart_upload,
                        multipart_upload_allowed=multipart_upload_allowed,
                    )
        else:
            if rpath.endswith("/"):
                rpath = os.path.join(rpath, local_path.name)
            _add_file_for_upload(
                local_file=str(lpath),
                remote_path=str(rpath),
                files_for_normal_upload=files_for_normal_upload,
                files_for_multipart_upload=files_for_multipart_upload,
                multipart_upload_allowed=multipart_upload_allowed,
            )
        return self._upload(
            files_for_normal_upload=files_for_normal_upload,
            files_for_multipart_upload=files_for_multipart_upload,
        )

    def _upload(
        self,
        files_for_normal_upload: List[Tuple[str, str, _FileMultiPartInfo]],
        files_for_multipart_upload: List[Tuple[str, str, _FileMultiPartInfo]],
    ):
        abort_event = Event()
        with ThreadPoolExecutor(
            max_workers=ENV_VARS.TFY_ARTIFACTS_UPLOAD_MAX_WORKERS
        ) as executor:
            futures: List[Future] = []
            for remote_path, local_path, _ in files_for_normal_upload:
                futures.append(
                    executor.submit(
                        self.client.upload,
                        file_path=local_path,
                        storage_uri=remote_path,
                    )
                )

            done, not_done = wait(futures, return_when=FIRST_EXCEPTION)
            if len(not_done) > 0:
                abort_event.set()
            for future in not_done:
                future.cancel()
            for future in done:
                if future.exception() is not None:
                    raise future.exception()

            for remote_path, local_path, multipart_info in files_for_multipart_upload:
                self._multipart_upload(
                    local_file=local_path,
                    artifact_path=remote_path,
                    multipart_info=multipart_info,
                    executor=executor,
                    abort_event=abort_event,
                )

    def _multipart_upload(
        self,
        local_file: str,
        artifact_path: str,
        multipart_info: _FileMultiPartInfo,
        executor: ThreadPoolExecutor,
        abort_event: Optional[Event] = None,
    ):
        logger.info(
            "Uploading %s to %s using multipart upload",
            local_file,
            artifact_path,
        )

        multipart_upload: SignedURLMultipartUploadAPIResponseDto = (
            self.client.create_multipart_upload(
                storage_uri=artifact_path,
                num_parts=multipart_info.num_parts,
            )
        )
        s3_compatible_multipart_upload(
            multipart_upload=MultiPartUpload(
                storage_provider=multipart_upload.storageProvider,
                part_signed_urls=[
                    SignedURL(signed_url=url.signedUrl)
                    for url in multipart_upload.partSignedUrls
                ],
                s3_compatible_upload_id=multipart_upload.uploadId,
                finalize_signed_url=SignedURL(
                    signed_url=multipart_upload.finalizeSignedUrl
                ),
            ),
            local_file=local_file,
            executor=executor,
            multipart_info=multipart_info,
            abort_event=abort_event,
        )

    @log_time(prefix=LOG_PREFIX)
    def isdir(self, path):
        """Is this entry directory-like?"""
        return self.client.is_directory(path)

    def open(
        self,
        path,
        mode="rb",
        block_size=None,
        cache_options=None,
        compression=None,
        **kwargs,
    ):
        """
        Open a file for reading or writing.
        """
        if "r" in mode:
            # Reading mode
            file_content = self.client.download_to_bytes(path)
            return io.BytesIO(file_content)
        elif "w" in mode or "a" in mode:
            # Writing mode (appending treated as writing)
            buffer = io.BytesIO()
            buffer.seek(0)

            def on_close(buffer=buffer, path=path):
                """
                Callback when file is closed, automatically upload the content.
                """
                buffer.seek(0)
                self.client.upload_from_bytes(buffer.read(), storage_uri=path)

            # Wrapping the buffer to automatically upload on close
            return io.BufferedWriter(buffer, on_close)

    @log_time(prefix=LOG_PREFIX)
    def write(self, path, data, **kwargs):
        """
        Write data to the file at the specified path (this could be tied to open's close).
        """
        if isinstance(data, io.BytesIO):
            data.seek(0)
            content = data.read()
        elif isinstance(data, str):
            content = data.encode()  # Encode to bytes
        else:
            raise ValueError("Unsupported data type for writing")

        # Upload the content to the remote file system
        self.client.upload_from_bytes(content, storage_uri=path)

    @log_time(prefix=LOG_PREFIX)
    def ls(self, path, detail=True, **kwargs):
        """List objects at path."""
        return self.client.list_files(path, detail=detail)


class SignedURLBufferedFile(AbstractBufferedFile):
    """
    Buffered file implementation for Signed URL-based file system.
    # TODO: Need to test this implementation
    """

    def __init__(
        self, fs: SignedURLFileSystem, path: str, mode: str, block_size: int, **kwargs
    ):
        """
        Initialize the buffered file, determining the mode (read/write).
        """
        super().__init__(fs, path, mode, block_size, **kwargs)
        self.buffer = io.BytesIO()
        self.client = fs.client

        if "r" in mode:
            # Download the file content for reading
            file_content = fs.client.download_to_bytes(path)
            self.buffer.write(file_content)
            self.buffer.seek(0)  # Reset buffer after writing content

    def _upload_on_close(self):
        """
        Upload content back to the remote store when the file is closed.
        """
        self.buffer.seek(0)
        self.client.upload_from_bytes(self.buffer.read(), storage_uri=self.path)

    def close(self):
        """
        Close the file, ensuring the content is uploaded for write/append modes.
        """
        if self.writable():
            self._upload_on_close()
        self.buffer.close()
        super().close()

    def _fetch_range(self, start, end):
        """
        Fetch a specific byte range from the file. Useful for large files and range reads.
        """
        self.buffer.seek(start)
        return self.buffer.read(end - start)

    def _upload_chunk(self, final=False):
        """
        Upload a chunk of the file. For larger files, data may be uploaded in chunks.
        """
        if final:
            self._upload_on_close()
